@@ -9,11 +9,13 @@
 # error "use a 2.6 kernel, please"
 #endif
 
-#include <linux/bio.h>
-#ifndef bio_rw_flagged
-#define bio_rw_flagged(bio, flag)      ((bio)->bi_rw & (1 << (flag)))
+#include <linux/blkdev.h>
+#ifndef BLKDEV_IFL_WAIT
+/* see fbd9b09a177a481eda256447c881f014f29034fe */
+#define blkdev_issue_flush(b, gfpf, s, ifl)	blkdev_issue_flush(b, s)
 #endif
 
+#include <linux/bio.h>
 #include <linux/slab.h>
 
 /* for the proc_create wrapper */
@@ -218,6 +220,24 @@ static inline void drbd_plug_device(struct drbd_conf *mdev)
 	spin_unlock_irq(q->queue_lock);
 }
 
+static inline int drbd_backing_bdev_events(struct drbd_conf *mdev)
+{
+	struct gendisk *disk = mdev->ldev->backing_bdev->bd_contains->bd_disk;
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15)
+	/* very old kernel */
+	return (int)disk_stat_read(disk, read_sectors)
+	     + (int)disk_stat_read(disk, write_sectors);
+#elif defined(__disk_stat_inc)
+	/* older kernel */
+	return (int)disk_stat_read(disk, sectors[0])
+	     + (int)disk_stat_read(disk, sectors[1]);
+#else
+	/* recent kernel */
+	return (int)part_stat_read(&disk->part0, sectors[0])
+	     + (int)part_stat_read(&disk->part0, sectors[1]);
+#endif
+}
+
 #ifdef DEFINE_SOCK_CREATE_KERN
 #define sock_create_kern sock_create
 #endif
@@ -694,6 +714,12 @@ static inline void blk_queue_max_segments(struct request_queue *q, unsigned shor
 #endif
 
 #ifdef NEED_ATOMIC_ADD_UNLESS
+#ifndef atomic_xchg
+static inline int atomic_xchg(atomic_t *v, int new)
+{
+	return xchg(&v->counter, new);
+}
+#endif
 #ifndef atomic_cmpxchg
 static inline int atomic_cmpxchg(atomic_t *v, int old, int new)
 {
@@ -730,4 +756,63 @@ static inline int atomic_add_unless(atomic_t *v, int a, int u)
 typedef _Bool                   bool;
 #endif
 
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,35)
+#ifdef REQ_HARDBARRIER
+#undef REQ_HARDBARRIER
+#endif
+#define REQ_HARDBARRIER (1UL << BIO_RW_BARRIER)
+#ifdef BIO_RW_SYNC
+/* see upstream commits
+ * 213d9417fec62ef4c3675621b9364a667954d4dd,
+ * 93dbb393503d53cd226e5e1f0088fe8f4dbaa2b8
+ * later, the defines even became an enum ;-) */
+#define REQ_SYNC        (1UL << BIO_RW_SYNC)
+#define REQ_UNPLUG      (0)
+#define DP_BCOMP_UNPLUG DP_UNPLUG
+#define REQ_BCOMP_SYNC  REQ_SYNC
+#else
+#define REQ_SYNC        (1UL << BIO_RW_SYNCIO)
+#define REQ_UNPLUG      (1UL << BIO_RW_UNPLUG)
+#define DP_BCOMP_UNPLUG (0)
+#define REQ_BCOMP_SYNC  (0)
+#endif
+#endif
+
+#ifndef REQ_FLUSH
+#define REQ_FLUSH       (0)
+#endif
+
+#ifndef REQ_FUA
+#define REQ_FUA		(REQ_HARDBARRIER|REQ_UNPLUG)
+#endif
+
+#ifndef REQ_DISCARD
+#define REQ_DISCARD     (0)
+#endif
+
+#ifdef NEED_SCHEDULE_TIMEOUT_INTERR
+static inline signed long schedule_timeout_interruptible(signed long timeout)
+{
+	__set_current_state(TASK_INTERRUPTIBLE);
+        return schedule_timeout(timeout);
+}
+
+static inline signed long schedule_timeout_uninterruptible(signed long timeout)
+{
+        __set_current_state(TASK_UNINTERRUPTIBLE);
+        return schedule_timeout(timeout);
+}
+#endif
+
+#ifndef CONFIG_DYNAMIC_DEBUG
+/* At least in 2.6.34 the function macro dynamic_dev_dbg() is broken when compiling
+   without CONFIG_DYNAMIC_DEBUG. It has 'format' in the argument list, it references
+   to 'fmt' in its body. */
+#ifdef dynamic_dev_dbg
+#undef dynamic_dev_dbg
+#define dynamic_dev_dbg(dev, fmt, ...)                               \
+        do { if (0) dev_printk(KERN_DEBUG, dev, fmt, ##__VA_ARGS__); } while (0)
+#endif
+#endif
+
 #endif
