@@ -30,15 +30,12 @@ extern const char *drbd_buildtag(void);
 #undef PRO_VERSION_MAX
 #endif
 
-#ifndef DRBD_DEBUG_STATE_CHANGES
-#define DRBD_DEBUG_STATE_CHANGES 0
-#endif
 /* End of external module for 2.6.33 stuff */
 
-#define REL_VERSION "8.3.8.1"
+#define REL_VERSION "8.3.9"
 #define API_VERSION 88
 #define PRO_VERSION_MIN 86
-#define PRO_VERSION_MAX 94
+#define PRO_VERSION_MAX 95
 
 #ifndef __CHECKER__   /* for a sparse run, we need all STATICs */
 #define DBG_ALL_SYMBOLS /* no static functs, improves quality of OOPS traces */
@@ -126,4 +123,7 @@ extern const char *drbd_buildtag(void);
 /* some old kernels do not have the bool type */
 //#define NEED_BOOL_TYPE
 
+/* some older kernels do not have schedule_timeout_interruptible() */
+//#define NEED_SCHEDULE_TIMEOUT_INTERR
+
 #endif
