@@ -233,11 +233,29 @@ void range_check(const enum range_checks what, const char *name,
 				DRBD_OUTDATED_WFC_TIMEOUT_MAX);
 		break;
 
-	case R_DP_VOLUME:
-	case R_DP_INTERVAL:
-	case R_RS_THROTTLE_TH:
-	case R_RS_HOLD_OFF_TH:
-		/* ignored anyways; emit deprecation warning? */
+	case R_C_PLAN_AHEAD:
+		m_strtoll_range(value, 1, name, DRBD_C_PLAN_AHEAD_MIN,
+				DRBD_C_PLAN_AHEAD_MAX);
+		break;
+
+	case R_C_DELAY_TARGET:
+		m_strtoll_range(value, 1, name, DRBD_C_DELAY_TARGET_MIN,
+				DRBD_C_DELAY_TARGET_MAX);
+		break;
+
+	case R_C_FILL_TARGET:
+		m_strtoll_range(value, 's', name, DRBD_C_FILL_TARGET_MIN,
+				DRBD_C_FILL_TARGET_MAX);
+		break;
+
+	case R_C_MAX_RATE:
+		m_strtoll_range(value, 'k', name, DRBD_C_MAX_RATE_MIN,
+				DRBD_C_MAX_RATE_MAX);
+		break;
+
+	case R_C_MIN_RATE:
+		m_strtoll_range(value, 'k', name, DRBD_C_MIN_RATE_MIN,
+				DRBD_C_MIN_RATE_MAX);
 		break;
 	}
 }
@@ -600,6 +618,9 @@ static struct d_option *parse_options_d(int token_switch, int token_option,
 		} else if (token == token_delegate) {
 			delegate(ctx);
 			continue;
+		} else if (token == TK_DEPRECATED_OPTION) {
+			/* fprintf(stderr, "Warn: Ignoring deprecated option '%s'\n", yylval.txt); */
+			expect_STRING_or_INT();
 		} else if (token == '}') {
 			return options;
 		} else {
