@@ -46,10 +46,11 @@ enum range_checks
 	R_WFC_TIMEOUT,
 	R_DEGR_WFC_TIMEOUT,
 	R_OUTDATED_WFC_TIMEOUT,
-	R_DP_VOLUME,
-	R_DP_INTERVAL,
-	R_RS_THROTTLE_TH,
-	R_RS_HOLD_OFF_TH,
+	R_C_PLAN_AHEAD,
+	R_C_DELAY_TARGET,
+	R_C_FILL_TARGET,
+	R_C_MAX_RATE,
+	R_C_MIN_RATE,
 };
 
 enum yytokentype {
@@ -113,6 +114,7 @@ enum yytokentype {
 	TK_NET_DELEGATE,
 	TK_INCLUDE,
 	TK_FLOATING,
+	TK_DEPRECATED_OPTION,
 };
 
 typedef struct YYSTYPE {
