@@ -203,18 +203,6 @@ static void show_handler(struct drbd_option *od, unsigned short* tp);
 static void show_bit(struct drbd_option *od, unsigned short* tp);
 static void show_string(struct drbd_option *od, unsigned short* tp);
 
-// ignore things that used to be options, but are no-ops now
-static int conv_ignore(struct drbd_option *od, struct drbd_tag_list *tl, char* arg)
-	{ return NO_ERROR; };
-#if 0
-static void show_ignore(struct drbd_option *od, unsigned short* tp)
-	{ };
-# endif
-static int usage_ignore(struct drbd_option *option, char* str, int strlen)
-	{ return 0; };
-static void usage_xml_ignore(struct drbd_option *option)
-	{ };
-
 // sub functions for events_cmd
 static int print_broadcast_events(unsigned int seq, int, struct drbd_nl_cfg_reply *reply);
 static int w_connected_state(unsigned int seq, int, struct drbd_nl_cfg_reply *reply);
@@ -262,6 +250,11 @@ const char *rrcf_n[] = {
 	[ASB_CALL_HELPER]        = "call-pri-lost"
 };
 
+const char *on_no_data_n[] = {
+	[OND_IO_ERROR]		= "io-error",
+	[OND_SUSPEND_IO]	= "suspend-io"
+};
+
 struct option wait_cmds_options[] = {
 	{ "wfc-timeout",required_argument, 0, 't' },
 	{ "degr-wfc-timeout",required_argument,0,'d'},
@@ -274,11 +267,6 @@ struct option wait_cmds_options[] = {
 	conv_numeric, show_numeric, numeric_opt_usage, numeric_opt_xml, \
 	{ .numeric_param = { DRBD_ ## N ## _MIN, DRBD_ ## N ## _MAX, \
 		DRBD_ ## N ## _DEF ,U,UN  } }
-/* deprecated numeric, only present in "drbdsetup show", otherwise ignored. */
-#define DN(N,U,UN) \
-	conv_ignore, show_numeric, usage_ignore, usage_xml_ignore, \
-	{ .numeric_param = { DRBD_ ## N ## _MIN, DRBD_ ## N ## _MAX, \
-		DRBD_ ## N ## _DEF ,U,UN  } }
 #define EN_sndbuf(N,U,UN) \
 	conv_sndbuf, show_numeric, numeric_opt_usage, numeric_opt_xml, \
 	{ .numeric_param = { DRBD_ ## N ## _MIN, DRBD_ ## N ## _MAX, \
@@ -373,13 +361,12 @@ struct drbd_cmd commands[] = {
 		 { "verify-alg", 'v',T_verify_alg,      ES },
 		 { "cpu-mask",'c',T_cpu_mask,           ES },
 		 { "use-rle",'R',T_use_rle,   EB },
-		 /* deprecated tags are hardcoded here, as they are no longer
-		  * used in current DRBD and would otherwise unnecessarily
-		  * clutter the generated struct definition */
-		 { "delay-probe-volume",'V',	71,	DN(DP_VOLUME,'k',"bytes") },
-		 { "delay-probe-interval",'I',	72,	DN(DP_INTERVAL,1,"1/10 seconds") },
-		 { "throttle-threshold",'T',	73,	DN(RS_THROTTLE_TH,1,"1/10 seconds") },
-		 { "hold-off-threshold",'H',	74,	DN(RS_HOLD_OFF_TH,1,"1/10 seconds") },
+		 { "on-no-data-accessible",'n',	T_on_no_data, EH(on_no_data_n,ON_NO_DATA) },
+		 { "c-plan-ahead", 'p',         T_c_plan_ahead, EN(C_PLAN_AHEAD,1,"1/10 seconds") },
+		 { "c-delay-target", 'd',       T_c_delay_target, EN(C_DELAY_TARGET,1,"1/10 seconds") },
+		 { "c-fill-target", 's',        T_c_fill_target, EN(C_FILL_TARGET,'s',"bytes") },
+		 { "c-max-rate", 'M',		T_c_max_rate, EN(C_MAX_RATE,'k',"bytes/second") },
+		 { "c-min-rate", 'm',	        T_c_min_rate, EN(C_MIN_RATE,'k',"bytes/second") },
 		 CLOSE_OPTIONS }} }, },
 
 	{"new-current-uuid", P_new_c_uuid, F_CONFIG_CMD, {{NULL,
@@ -487,6 +474,7 @@ static const char *error_messages[] = {
 	EM(ERR_CSUMS_RESYNC_RUNNING) = "Can not change csums-alg while resync is in progress",
 	EM(ERR_PERM) = "Permission denied. CAP_SYS_ADMIN necessary",
 	EM(ERR_NEED_APV_93) = "Protocol version 93 required to use --assume-clean",
+	EM(ERR_STONITH_AND_PROT_A) = "Fencing policy resource-and-stonith only with prot B or C allowed",
 };
 #define MAX_ERROR (sizeof(error_messages)/sizeof(*error_messages))
 const char * error_to_string(int err_no)
@@ -640,22 +628,39 @@ static int conv_md_idx(struct drbd_argument *ad, struct drbd_tag_list *tl, char*
 	return NO_ERROR;
 }
 
-static void resolv6(char *name, struct in6_addr *addr)
+static void resolv6(char *name, struct sockaddr_in6 *addr)
 {
-	int rv;
-	struct hostent *he;
+	struct addrinfo hints, *res, *tmp;
+	int err;
 
-	rv = inet_pton(AF_INET6, name, addr);
-	if (rv > 0)
-		return;
-	else if (rv == 0) {
-		he = gethostbyname2(name, AF_INET6);
-		if (!he) {
-			fprintf(stderr, "can not resolve the hostname: gethostbyname2(%s, AF_INET6): %s\n",
-					name, hstrerror(h_errno));
-			exit(20);
-		}
-		memcpy(addr, he->h_addr_list[0], sizeof(struct in6_addr));
+	memset(&hints, 0, sizeof(hints));
+	hints.ai_family = AF_INET6;
+	hints.ai_socktype = SOCK_STREAM;
+	hints.ai_protocol = IPPROTO_TCP;
+
+	err = getaddrinfo(name, 0, &hints, &res);
+	if (err) {
+		fprintf(stderr, "getaddrinfo %s: %s\n", name, gai_strerror(err));
+		exit(20);
+	}
+
+	/* Yes, it is a list. We use only the first result. The loop is only
+	 * there to document that we know it is a list */
+	for (tmp = res; tmp; tmp = tmp->ai_next) {
+		memcpy(addr, tmp->ai_addr, sizeof(*addr));
+		break;
+	}
+	freeaddrinfo(res);
+	if (0) { /* debug output */
+		char ip[INET6_ADDRSTRLEN];
+		inet_ntop(AF_INET6, &addr->sin6_addr, ip, sizeof(ip));
+		fprintf(stderr, "%s -> %02x %04x %08x %s %08x\n",
+				name,
+				addr->sin6_family,
+				addr->sin6_port,
+				addr->sin6_flowinfo,
+				ip,
+				addr->sin6_scope_id);
 	}
 }
 
@@ -676,6 +681,25 @@ static unsigned long resolv(const char* name)
 	return retval;
 }
 
+static void split_ipv6_addr(char **address, int *port)
+{
+	/* ipv6:[fe80::0234:5678:9abc:def1]:8000; */
+	char *b = strrchr(*address,']');
+	if (address[0][0] != '[' || b == NULL ||
+		(b[1] != ':' && b[1] != '\0')) {
+		fprintf(stderr, "unexpected ipv6 format: %s\n",
+				*address);
+		exit(20);
+	}
+
+	*b = 0;
+	*address += 1; /* skip '[' */
+	if (b[1] == ':')
+		*port = m_strtoll(b+2,1); /* b+2: "]:" */
+	else
+		*port = 7788; /* will we ever get rid of that default port? */
+}
+
 static void split_address(char* text, int *af, char** address, int* port)
 {
 	static struct { char* text; int af; } afs[] = {
@@ -697,15 +721,27 @@ static void split_address(char* text, int *af, char** address, int* port)
 			break;
 		}
 	}
+
+	if (*af == AF_INET6 && address[0][0] == '[')
+		return split_ipv6_addr(address, port);
+
 	if (*af == -1)
 		*af = get_af_ssocks(1);
 
 	b=strrchr(text,':');
 	if (b) {
 		*b = 0;
+		if (*af == AF_INET6) {
+			/* compatibility handling of ipv6 addresses,
+			 * in the style expected before drbd 8.3.9.
+			 * may go wrong without explicit port */
+			fprintf(stderr, "interpreting ipv6:%s:%s as ipv6:[%s]:%s\n",
+					*address, b+1, *address, b+1);
+		}
 		*port = m_strtoll(b+1,1);
 	} else
 		*port = 7788;
+
 }
 
 static int conv_address(struct drbd_argument *ad, struct drbd_tag_list *tl, char* arg)
@@ -729,13 +765,12 @@ static int conv_address(struct drbd_argument *ad, struct drbd_tag_list *tl, char
 
 	if (af == AF_INET6) {
 		memset(&addr6, 0, sizeof(struct sockaddr_in6));
+		resolv6(address, &addr6);
 		addr6.sin6_port = htons(port);
-		addr6.sin6_family = AF_INET6;
-		resolv6(address, &addr6.sin6_addr);
-		/* addr6.sin6_len = sizeof(addr6); */
 		add_tag(tl,ad->tag,&addr6,sizeof(addr6));
 	} else {
-		/* AF_INET and AF_SSOCKS */
+		/* AF_INET, AF_SDP, AF_SSOCKS,
+		 * all use the IPv4 addressing scheme */
 		addr.sin_port = htons(port);
 		addr.sin_family = af;
 		addr.sin_addr.s_addr = resolv(address);
